<?php

require_once(dirname(__FILE__)."/../mapviewerphp/constants.php");

$sessionId = $_REQUEST["SESSION"];
$mapName = $_REQUEST["MAPNAME"];
$sourceId = $_REQUEST["SOURCE_ID"];
$targetId = $_REQUEST["TARGET_ID"];
$layerName = "Results";

$scName = "Default";
$idPropName = "ID";
$geomPropName = "Geom";
$schemaName = "Default";
$className = "ShortestPath";
$classNameQualified = "$schemaName:$className";
$fsIdStr = "Session:$sessionId//ShortestPath.FeatureSource";
$layerIdStr = "Session:$sessionId//ShortestPath.LayerDefinition";

//This function creates the Shortest Path results layer if it doesn't already
//exist. If it does, it returns that instead.
function CreateResultsLayer($sessionId, $map, $featSvc, $resSvc) {
    global $fsIdStr;
    global $layerIdStr;
    global $schemaName;
    global $className;
    global $classNameQualified;
    global $idPropName;
    global $geomPropName;
    global $scName;
    global $layerName;

    //Our results layer has a simple schema
    // ID: Road ID
    // Geom: Geometry of the road segment
    $layers = $map->GetLayers();
    $layerIdx = $layers->IndexOf($layerName);
    if ($layerIdx >= 0) {
        return $layers->GetItem($layerIdx);
    } else {
        //Set up the results schema
        $fsId = new MgResourceIdentifier($fsIdStr);
        $layerId = new MgResourceIdentifier($layerIdStr);
        
        $csFact = new MgCoordinateSystemFactory();
        //Use the same CS as the roads (EPSG:4617)
        $wkt = $csFact->ConvertEpsgCodeToWkt(4617);
        
        $schema = new MgFeatureSchema($schemaName, "pgRouting results schema");
        $klass = new MgClassDefinition();
        $klass->SetName($className);
        
        $clsProps = $klass->GetProperties();
        $idProps = $klass->GetIdentityProperties();
        
        $idProp = new MgDataPropertyDefinition($idPropName);
        $idProp->SetDataType(MgPropertyType::Int64);
        $idProp->SetAutoGeneration(false);
        $idProp->SetNullable(false);
        
        $geomProp = new MgGeometricPropertyDefinition($geomPropName);
        $geomProp->SetSpatialContextAssociation($scName);
        $geomProp->SetGeometryTypes(MgFeatureGeometricType::Point | MgFeatureGeometricType::Curve | MgFeatureGeometricType::Surface);
        
        $clsProps->Add($idProp);
        $idProps->Add($idProp);
        
        $clsProps->Add($geomProp);
        $klass->SetDefaultGeometryPropertyName($geomPropName);
        
        $classes = $schema->GetClasses();
        $classes->Add($klass);
        
        //Create the SDF feature source
        $createFs = new MgFileFeatureSourceParams("OSGeo.SDF", $scName, $wkt, $schema);
        $featSvc->CreateFeatureSource($fsId, $createFs);
        
        //Now set up a layer definition
        $content = file_get_contents(dirname(__FILE__)."/layerdef.xml");
        $content = sprintf($content, $fsIdStr, $classNameQualified, $geomPropName);
        $bs = new MgByteSource($content, strlen($content));
        $br = $bs->GetReader();
        $resSvc->SetResource($layerId, $br, NULL);
        
        //Make a runtime layer from it
        $resultsLayer = new MgLayer($layerId, $resSvc);
        
        //Insert to top of draw order
        $layers->Insert(0, $resultsLayer);
        
        $resultsLayer->SetName($layerName);
        $resultsLayer->SetDisplayInLegend(true);
        $resultsLayer->SetLegendLabel("Shortest Path Query Result");
        $resultsLayer->SetSelectable(false);
        
        //All good to go
        return $resultsLayer;
    }
}

//This function loads the contents of the MgSqlDataReader into our
//Shortest Path results layer.
function LoadResults($layer, $reader) {
    global $idPropName;
    global $geomPropName;
    global $classNameQualified;

    $commands = new MgFeatureCommandCollection();
    $batchProps = new MgBatchPropertyCollection();
    
    //LEAKY ABSTRACTION:
    //
    //The PostgreSQL provider is upper-casing the column names in the SQL reader we get
    //back even though our query wanted these as lower-case.
    $idCol = "GID";
    $geomCol = "GEOM";
    
    //Loop through each record in the SQL reader and load it into a MgBatchPropertyCollection. The MgInsertFeatures
    //can accept a MgPropertyCollection (a single feature to be inserted) or a MgBatchPropertyCollection
    //(multiple features to be inserted). We're using the latter method.
    while ($reader->ReadNext()) {
        //Not sure what data type gid will come back as, so test for the common expected types
        $pt = $reader->GetPropertyType($idCol);
        $id = -1;
        if ($pt == MgPropertyType::Int64) {
            $id = $reader->GetInt64($idCol); 
        } else {
            $id = $reader->GetInt32($idCol); 
        }
        //This is why this sample is possible, the "geom" column comes back as AGF binary that we
        //can then either create a MgGeometry out of, or in the case of this sample, to be fed
        //into our temporary feature source.
        $agf = $reader->GetGeometry($geomCol);
        
        $propVals = new MgPropertyCollection();
        $idVal = new MgInt64Property($idPropName, $id);
        $geomVal = new MgGeometryProperty($geomPropName, $agf);
        
        $propVals->Add($idVal);
        $propVals->Add($geomVal);
        
        $batchProps->Add($propVals);
    }
    $reader->Close();
    
    //Set up our commands. They will be executed sequentially in the order specified. So we are
    // 1. Deleting everything that's in our feature source
    // 2. Inserting our new shortest path results
    $deleteCmd = new MgDeleteFeatures($classNameQualified, "$idPropName >= -1");
    $commands->Add($deleteCmd);
    $insertCmd = new MgInsertFeatures($classNameQualified, $batchProps);
    $commands->Add($insertCmd);
    
    $errorDetails = NULL;
    $insertedFeatures = 0;
    
    //Call MgLayer::UpdateFeatures(), we can do this because we created a MgMap with MgMap(MgSiteConneciton)
    //and not the *deprecated* MgMap() constructor.
    $results = $layer->UpdateFeatures($commands);
    for ($i = 0; $i < $results->GetCount(); $i++) {
        $resultVal = $results->GetItem($i);
        //Good practice: Close any feature readers
        if ($resultVal->GetPropertyType() == MgPropertyType::Feature) {
            $fr = $resultVal->GetValue();
            while ($fr->ReadNext()) {
                $insertedFeatures++;
            }
            $fr->Close();
        }
        //Also good practice: Log/aggregate any error messages
        else if ($resultVal->GetPropertyType() == MgPropertyType::String) {
            $errorDetails = $resultVal->GetValue();
        }
    }
    
    if ($errorDetails != NULL) {
        throw new Exception($errorDetails);
    }
    
    return $insertedFeatures;
}

try {
    MgInitializeWebTier(dirname(__FILE__)."/../webconfig.ini");
    $siteConn = new MgSiteConnection();
    $userInfo = new MgUserInformation($sessionId);
    $siteConn->Open($userInfo);
    
    //Note we're not using "new MgMap()", means we can access convenience APIs
    $map = new MgMap($siteConn);
    $map->Open($mapName);
    
    $featSvc = $siteConn->CreateService(MgServiceType::FeatureService);
    $resSvc = $siteConn->CreateService(MgServiceType::ResourceService);
    
    //This is our PostGIS feature source
    $fsId = new MgResourceIdentifier("Library://pg_routing/Data/postgis.FeatureSource");
    
    //Create the shortest path results layer if necessary
    $resultsLayer = CreateResultsLayer($sessionId, $map, $featSvc, $resSvc, $layerName);
    
    //Execute our pgRouting SQL query. The shortest_path function itself is not spatial, so we
    //need to join the results of that query back into the roads table to get the associated line segments
    //that comprise the shortest path
    //
    //The key part that makes this sample all work is that the "geom" column inside the returning MgSqlDataReader
    //is an actual FDO geometry value and not some unreadable BLOB.
    //
    //Now although we can't directly render a layer off of a SQL query (wouldn't that be nice to have?), what we
    //can do instead is pump the reader results into a temporary feature source and make a layer off of that. The
    //shortest path results layer is our layer for this example.
    
    //This is the inner SQL query we need to pass to shortest_path()
    $innerSql = "select gid as id, source::integer, target::integer, length::double precision as cost from roads";
    
    //This is the SQL query that joins the results of the shortest_path() query with the roads table to get the
    //associated geometries that comprise our shortest path
    $sql = "select gid, geom from roads 
            join (select * from shortest_path('$innerSql', $sourceId, $targetId, false, false)) as route 
            on roads.gid = route.edge_id";
    
    $reader = $featSvc->ExecuteSqlQuery($fsId, $sql);
    
    //Load this query result into the results layer. We can do this because the PostgreSQL FDO
    //provider is correctly interpeting any geometry results as FDO geometry values, which is what
    //we need for this whole thing to a) be feasible b) work
    $count = LoadResults($resultsLayer, $reader);
    
    //Make sure to save our changes
    $resultsLayer->ForceRefresh();
    $map->Save();
} catch (MgException $ex) {
    echo "<strong>MapGuide Error:</strong><pre>".$ex->GetDetails()."</pre>";
    echo "<a href='index.php?SESSION=$sessionId&MAPNAME=$mapName'>Return</a>";
    die;
} catch (Exception $e) {
    echo "<strong>Error:</strong><pre>".$e->getMessage()."</pre>";
    echo "<a href='index.php?SESSION=$sessionId&MAPNAME=$mapName'>Return</a>";
    die;
}
?>
<html>
    <head>
        <title>Shortest Path Query Result</title>
        <script type="text/javascript" src="common.js"></script>
        <script type="text/javascript">
        <? if ($count > 0) { ?>
        
        window.onload = function() {
            if (isFusion()) {
                var Fusion = getFusionInstance();
                var map = Fusion.getMapByIndice(0);
                var allLayers = map.getAllMaps();
                var mgLayer = null;
                for (var i = 0; i < allLayers.length; i++) {
                    if (allLayers[i].arch == "MapGuide") {
                        mgLayer = allLayers[i];
                        break;
                    }
                }
                //This triggers rebuild of legend tree
                if (mgLayer != null) {
                    mgLayer.reloadMap();
                    mgLayer.drawMap();
                }
            } else {
                var viewer = getViewerInstance();
                viewer.Refresh();
            }
        };
        
        <? } ?>
        </script>
    </head>
    <body>
        <? if ($count > 0) { ?>
        <p>Shortest path result updated</p>
        <? } else { ?>
        <p>Shortest path query did not return a result</p>
        <? } ?>
        <a href='index.php?SESSION=<?= $sessionId ?>&MAPNAME=<?= $mapName ?>'>Return</a>
        <p>DEBUG (pgRouting SQL):</p>
        <pre>
        <?= $sql ?>
        </pre>
    </body>
</html>